//! Stream buffering for sync of encrypted socked.
use super::*;
use smallvec::SmallVec;
use std::io;
use std::{
    task::{
	Poll, Context,
    },
    pin::Pin,
};
use bytes::{
    Buf, BufMut,
};

/// A wrapping buffer over a writer and/or reader.
#[pin_project]
#[derive(Debug)]
pub struct Buffered<T: ?Sized, const SIZE: usize>
{
    /// Current internal buffer
    /// When it's full to `SIZE`, it should be written to `stream` at once then cleared when it's been written.
    buffer: SmallVec<[u8; SIZE]>, //TODO: Can we have a non-spilling stack vec?
    #[pin] stream: T
}

impl<T, const SIZE: usize> Buffered<T, SIZE>
where [(); SIZE]: Sized, // This isn't checked?
{
    /// Create a new staticly sized buffer wrapper around this stream
    pub fn new(stream: T) -> Self
    {
	assert!(SIZE > 0, "Size of buffer cannot be 0");
	Self {
	    buffer: SmallVec::new(),
	    stream,
	}
    }
    /// Consume into the wrapped stream
    pub fn into_inner(self) -> T
    {
	self.stream
    }
    /// The inner stream
    pub fn inner(&self) -> &T
    {
	&self.stream
    }
    /// A mutable reference to the backing stream
    pub fn inner_mut(&mut self) -> &mut T
    {
	&mut self.stream
    }

    /// The current buffer bytes.
    pub fn current_buffer(&self) -> &[u8]
    {
	&self.buffer[..]
    }

    /// Is the current internal buffer empty?
    ///
    /// You can flush a partially-filled buffer to the backing stream of a writer with `.flush().await`.
    pub fn is_empty(&self) -> bool
    {
	self.buffer.is_empty()
    }
}

#[inline] fn div_mod<V>(a: V, b: V) -> (V, <V as std::ops::Div>::Output, <V as std::ops::Rem>::Output)
where V: std::ops::Div + std::ops::Rem + Clone
{
    (a.clone(), a.clone() / b.clone(), a % b)
}

// XXX: I don't think writing futures like this is safe. Expand the inline `async{}`s into actual polling.
impl<T: ?Sized + Unpin, const SIZE: usize> AsyncWrite for Buffered<T, SIZE>
where T: AsyncWrite
{
    fn poll_write(self: Pin<&mut Self>, cx: &mut Context<'_>, buf: &[u8]) -> Poll<Result<usize, io::Error>> {
	// TODO: Don't write poll methods like this ffs... Write it properly.
	let this = self.get_mut();
	let fut = async {
	    use tokio::prelude::*;
	    
	    let mut written=0;
	    let mut err = None;
	    this.buffer.extend_from_slice(buf);

	    for chunk in this.buffer.chunks_exact(SIZE)
	    {
		if cfg!(test) {
		    eprintln!("Pushing chunk: {:?}", chunk);
		}
		match this.stream.write_all(&chunk).await {
		    Ok(()) => {
			written += chunk.len();
		    },
		    Err(e) => {
			err = Some(e);
			break;
		    },
		}
	    }
	    this.buffer.drain(0..written);
	    if let Some(err) = err {
		Err(err)
	    } else {
		Ok(buf.len())
	    }
	};
	tokio::pin!(fut);
	fut.poll(cx)
	    
    }
    fn poll_flush(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
	let this = self.get_mut();
	let fut = async {
	    use tokio::prelude::*;
	    
	    let wres = if this.buffer.len() > 0 {
		if cfg!(test) {
		    eprintln!("Pushing rest: {:?}", &this.buffer[..]);
		}
		let res = this.stream.write_all(&this.buffer[..]).await;
		this.buffer.clear();
		res
	    } else {
		Ok(())
	    };
	    this.stream.flush().await?;
	    wres
	};
	tokio::pin!(fut);
	fut.poll(cx)
    }
    fn poll_shutdown(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
	let this = self.get_mut();
	let fut = async {
	    use tokio::prelude::*;
	    
	    this.flush().await?;
	    this.stream.shutdown().await
	};
	tokio::pin!(fut);
	fut.poll(cx)
    }
}

fn try_release_buf<const SIZE: usize>(buffer: &mut SmallVec<[u8; SIZE]>, into: &mut [u8]) -> (bool, usize)
{
    let sz = std::cmp::min(buffer.len(), into.len());
    (&mut into[..sz]).copy_from_slice(&buffer[..sz]);
    drop(buffer.drain(..sz));
    (!buffer.is_empty(), sz)
}

// XXX: I don't think writing futures like this is safe. Expand the inline `async{}`s into actual polling.
impl<T: ?Sized + Unpin, const SIZE: usize> AsyncRead for Buffered<T, SIZE>
where T: AsyncRead
{
    fn poll_read(self: Pin<&mut Self>, cx: &mut Context<'_>, buf: &mut [u8]) -> Poll<io::Result<usize>> {
	let this = self.get_mut();
	//XXX: Why TF doesn't this work!??!?!??!?!?
	let fut = async {
	    use tokio::prelude::*;

	    let mut w = 0;
	    let mut temp = [0u8; SIZE];
	    for buf in buf.chunks_mut(SIZE) {
		let mut done=0;
		let mut r;
		if this.buffer.len() == 0
		{
		    while done < SIZE {
			// Fill the buffer
			r = this.stream.read(&mut temp[done..]).await?;
			done += r;
			if r == 0 {
			    break;
			}
		    }
		    
		    this.buffer.extend_from_slice(&temp[..done]);
		}
		// Drain buffer into `buf` (chunk)
		let sz = std::cmp::min(this.buffer.len(), buf.len());
		
		println!("Buffer size: {}, {} -> {} ({})", this.buffer.len(), sz, buf.len(), w);
		(&mut buf[..sz]).copy_from_slice(&this.buffer[..sz]);
		drop(this.buffer.drain(0..sz));
		w+= sz;
		// Wtf?
		if done == 0 {
		    break;
		}
	    }
	    
	    Result::<usize, io::Error>::Ok(w)
	};
	tokio::pin!(fut);
	fut.poll(cx)
    }
}

#[cfg(test)]
mod tests
{
    use super::*;
    #[tokio::test]
    async fn writer() -> eyre::Result<()>
    {
	use tokio::prelude::*;
	
	let (tx, mut rx) = tokio::io::duplex(11);
	let mut ttx = Buffered::<_, 4>::new(tx);

	let back = tokio::spawn(async move {

	    println!("Writing bytes");
	    ttx.write_all(b"Hello world").await?;
	    
	    println!("Waiting 1 second...");
	    tokio::time::delay_for(tokio::time::Duration::from_secs(1)).await;
	    println!("Flushing stream");
	    ttx.flush().await?;
	    ttx.shutdown().await?;
	    Result::<_, std::io::Error>::Ok(())
	});

	let mut output = Vec::new();
	println!("Reading full stream...");
	tokio::io::copy(&mut rx, &mut output).await?;
	println!("Waiting for background...");
	back.await.expect("Back panick")?;

	println!("Expected {:?}, got {:?}", b"Hello world", &output);
	assert_eq!(&output[..], b"Hello world");
	
	Ok(())
    }

    #[tokio::test]
    async fn reader() -> eyre::Result<()>
    {
	use tokio::prelude::*;

	const DATA: &'static [u8] = b"Hello world";
	
	let (mut tx, rx) = tokio::io::duplex(11);
	let mut rx = Buffered::<_, 4>::new(rx);

	let back = tokio::spawn(async move {
	    tx.write_all(DATA).await?;
	    tx.write_all(DATA).await?;
	    tx.flush().await?;

	    tx.shutdown().await?;
	    Result::<_, std::io::Error>::Ok(())
	});

	let mut output =Vec::new();
	tokio::io::copy(&mut rx, &mut output).await?;
	
	back.await.expect("Back panick")?;

	eprintln!("String: {}", String::from_utf8_lossy(&output[..]));
	assert_eq!(&output[..DATA.len()], &DATA[..]);
	assert_eq!(&output[DATA.len()..], &DATA[..]);
	
	Ok(())
    }
}

